"""
Tests for fetching patient metadata

"""
from unittest import TestCase, mock

from runeq.config import Config
from runeq.resources.client import GraphClient
from runeq.resources.patient import (
    Device, DeviceSet, Patient, PatientSet, get_all_devices, get_all_patients,
    get_device, get_patient, get_patient_devices
)


class TestPatient(TestCase):
    """
    Unit tests for the Patient class.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.mock_client = GraphClient(
            Config(
                client_key_id='test',
                client_access_key='config'
            )
        )

    def test_attributes(self):
        """
        Test attributes for an initialized Patient.

        """
        test_patient = Patient(
            id="p1",
            created_at=1629300943.9179766,
            name="patient1",
            devices=DeviceSet(
                [
                    Device(
                        id="d1",
                        patient_id="p1",
                        name="Percept",
                        created_at=1629300943.9179766,
                        device_type_id="dt1",
                    ),
                    Device(
                        id="d2",
                        patient_id="p1",
                        name="Apple Watch",
                        created_at=1629300943.9179766,
                        device_type_id="dt2",
                    )
                ]
            )
        )

        self.assertEqual("p1", test_patient.id)
        self.assertEqual(1629300943.9179766, test_patient.created_at)
        self.assertEqual("patient1", test_patient.name)

        actual_devices = test_patient.devices.to_list()
        self.assertEqual(2, len(actual_devices))

        self.assertEqual(
            {
                "id": "d1",
                "patient_id": "p1",
                "name": "Percept",
                "created_at": 1629300943.9179766,
                "device_type_id": "dt1",
            },
            actual_devices[0]
        )

        self.assertEqual(
            {
                "id": "d2",
                "patient_id": "p1",
                "name": "Apple Watch",
                "created_at": 1629300943.9179766,
                "device_type_id": "dt2",
            },
            actual_devices[1]
        )

        self.assertEqual(
            Device(
                id="d1",
                patient_id="p1",
                name="Percept",
                created_at=1629300943.9179766,
                device_type_id="dt1",
            ),
            test_patient.device("d1")
        )

    def test_repr(self):
        """
        Test __repr__

        """
        test_patient = Patient(
            id="p1",
            created_at=1629300943.9179766,
            name="Patient 1",
            devices=DeviceSet()
        )
        self.assertEqual(
            'Patient(id="p1", name="Patient 1")',
            repr(test_patient)
        )

    def test_normalize_id(self):
        """
        Test strip id of resource prefix if it exists.

        """
        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Patient.normalize_id("patient-d4b1c627bd464fe0a5ed940cc8e8e485")
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Patient.normalize_id("d4b1c627bd464fe0a5ed940cc8e8e485")
        )

    def test_denormalize_id(self):
        """
        Test add id resource prefix if it doesn't exist.

        """
        self.assertEqual(
            "patient-d4b1c627bd464fe0a5ed940cc8e8e485",
            Patient.denormalize_id("patient-d4b1c627bd464fe0a5ed940cc8e8e485")
        )

        self.assertEqual(
            "patient-d4b1c627bd464fe0a5ed940cc8e8e485",
            Patient.denormalize_id("d4b1c627bd464fe0a5ed940cc8e8e485")
        )

    def test_get_patient_basic(self):
        """
        Test get patient for specified patient_id.

        """
        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "patient": {
                    "id": "p1",
                    "created_at": 1630515986.9949625,
                    "name": "patient1",
                    "deviceList": {
                        "pageInfo": {
                            "endCursor": None
                        },
                        "devices": [
                            {
                                "id": "d1",
                                "name": "Percept",
                                "created_at": 1646685476.1367705,
                                "device_type": {
                                    "id": "dt1",
                                    "name": "percept"
                                },
                                "disabled": False,
                                "disabled_at": None,
                                "updated_at": 1646685485.9403558
                            },
                            {
                                "id": "d2",
                                "name": "Strive PD",
                                "created_at": 1646684177.194158,
                                "device_type": {
                                    "id": "dt2",
                                    "name": "strivestudy"
                                },
                                "disabled": False,
                                "disabled_at": None,
                                "updated_at": 1646684177.194158
                            },
                            {
                                "id": "d3",
                                "name": "Apple Watch",
                                "created_at": 1646684177.1942198,
                                "device_type": {
                                    "id": "dt3",
                                    "name": "watch"
                                },
                                "disabled": True,
                                "disabled_at": 1646684178.1942198,
                                "updated_at": 1646684178.1942198
                            },
                        ]
                    }
                }
            }
        ]

        test_patient = get_patient("patient-p1", client=self.mock_client)

        self.assertEqual(
            {
                "id": "p1",
                "created_at": 1630515986.9949625,
                "name": "patient1",
                "devices": [
                    {
                        "patient_id": "p1",
                        "name": "Percept",
                        "created_at": 1646685476.1367705,
                        "device_type_id": "dt1",
                        "disabled": False,
                        "disabled_at": None,
                        "updated_at": 1646685485.9403558,
                        "id": "d1"
                    },
                    {
                        "patient_id": "p1",
                        "name": "Strive PD",
                        "created_at": 1646684177.194158,
                        "device_type_id": "dt2",
                        "disabled": False,
                        "disabled_at": None,
                        "updated_at": 1646684177.194158,
                        "id": "d2"
                    },
                    {
                        "patient_id": "p1",
                        "name": "Apple Watch",
                        "created_at": 1646684177.1942198,
                        "device_type_id": "dt3",
                        "disabled": True,
                        "disabled_at": 1646684178.1942198,
                        "updated_at": 1646684178.1942198,
                        "id": "d3"
                    }
                ]
            },
            test_patient.to_dict(),
        )


    def test_get_patient_paginated(self):
        """
        Test get patient for specified patient_id, where the gql requests
        page through devices.

        """
        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "patient": {
                    "id": "p1",
                    "created_at": 1630515986.9949625,
                    "name": "patient1",
                    "deviceList": {
                        "pageInfo": {
                            "endCursor": "test_check_next"
                        },
                        "devices": [
                            {
                                "id": "d1",
                                "name": "Percept",
                                "created_at": 1646685476.1367705,
                                "device_type": {
                                    "id": "dt1",
                                },
                                "disabled": False,
                                "disabled_at": None,
                                "updated_at": 1646685485.9403558
                            }
                        ]
                    }
                }
            },
            {
                "patient": {
                    "id": "p1",
                    "created_at": 1630515986.9949625,
                    "name": "patient1",
                    "deviceList": {
                        "pageInfo": {
                            "endCursor": None
                        },
                        "devices": [
                            {
                                "id": "d2",
                                "name": "Strive PD",
                                "created_at": 1646684177.194158,
                                "device_type": {
                                    "id": "dt2",
                                },
                                "disabled": False,
                                "disabled_at": None,
                                "updated_at": 1646684177.194158
                            }
                        ]
                    }
                }
            }
        ]

        test_patient = get_patient("patient-p1", client=self.mock_client)

        self.assertEqual(
            {
                "id": "p1",
                "created_at": 1630515986.9949625,
                "name": "patient1",
                "devices": [
                    {
                        "patient_id": "p1",
                        "name": "Percept",
                        "created_at": 1646685476.1367705,
                        "device_type_id": "dt1",
                        "disabled": False,
                        "disabled_at": None,
                        "updated_at": 1646685485.9403558,
                        "id": "d1"
                    },
                    {
                        "patient_id": "p1",
                        "name": "Strive PD",
                        "created_at": 1646684177.194158,
                        "device_type_id": "dt2",
                        "disabled": False,
                        "disabled_at": None,
                        "updated_at": 1646684177.194158,
                        "id": "d2"
                    }
                ]
            },
            test_patient.to_dict(),
        )

    def test_get_all_patients_basic(self):
        """
        Test get patients for the initialized user.

        """
        test_patient1 = {
            "patient": {
                "id": "p1",
                "created_at": 1630515986.9949625,
                "name": "patient1",
                "deviceList": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "devices": [
                        {
                            "id": "d1",
                            "name": "Percept",
                            "created_at": 1646685476.1367705,
                            "device_type": {
                                "id": "dt1",
                            },
                            "disabled": False,
                            "disabled_at": None,
                            "updated_at": 1646685485.9403558
                        }
                    ]
                }
            }
        }

        test_patient2 = {
            "patient": {
                "id": "p2",
                "created_at": 1630515986.9949625,
                "name": "patient2",
                "deviceList": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "devices": [
                        {
                            "id": "d1",
                            "name": "Percept",
                            "created_at": 1646685476.1367705,
                            "device_type": {
                                "id": "dt1",
                            },
                            "disabled": False,
                            "disabled_at": None,
                            "updated_at": 1646685485.9403558
                        }
                    ]
                }
            }
        }


        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "org": {
                    "patientAccessList": {
                        "pageInfo": {
                            "endCursor": None
                        },
                        "patientAccess": [
                            test_patient1,
                            test_patient2
                        ]
                    }
                }
            }
        ]

        patients = get_all_patients(client=self.mock_client)

        self.assertEqual(
            [
                {
                    "id": "p1",
                    "created_at": 1630515986.9949625,
                    "name": "patient1",
                    "devices": [
                        {
                            "patient_id": "p1",
                            "name": "Percept",
                            "created_at": 1646685476.1367705,
                            "device_type_id": "dt1",
                            "disabled": False,
                            "disabled_at": None,
                            "updated_at": 1646685485.9403558,
                            "id": "d1"
                        }
                    ]
                },
                {
                    "id": "p2",
                    "created_at": 1630515986.9949625,
                    "name": "patient2",
                    "devices": [
                        {
                            "patient_id": "p2",
                            "name": "Percept",
                            "created_at": 1646685476.1367705,
                            "device_type_id": "dt1",
                            "disabled": False,
                            "disabled_at": None,
                            "updated_at": 1646685485.9403558,
                            "id": "d1"
                        }
                    ]
                }
            ],
            patients.to_list(),
        )


    def test_get_all_patients_paginated(self):
        """
        Test get patients for the initialized user, where the gql requests
        page through patients.

        """
        test_patient1 = {
            "patient": {
                "id": "p1",
                "created_at": 1630515986.9949625,
                "name": "patient1",
                "deviceList": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "devices": [
                        {
                            "id": "d1",
                            "name": "Percept",
                            "created_at": 1646685476.1367705,
                            "device_type": {
                                "id": "dt1",
                            },
                            "disabled": False,
                            "disabled_at": None,
                            "updated_at": 1646685485.9403558
                        }
                    ]
                }
            }
        }

        test_patient2 = {
            "patient": {
                "id": "p2",
                "created_at": 1630515986.9949625,
                "name": "patient2",
                "deviceList": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "devices": [
                        {
                            "id": "d1",
                            "name": "Percept",
                            "created_at": 1646685476.1367705,
                            "device_type": {
                                "id": "dt1",
                            },
                            "disabled": False,
                            "disabled_at": None,
                            "updated_at": 1646685485.9403558
                        }
                    ]
                }
            }
        }


        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "org": {
                    "patientAccessList": {
                        "pageInfo": {
                            "endCursor": "test_check_next"
                        },
                        "patientAccess": [
                            test_patient1
                        ]
                    }
                }
            },
            {
                "org": {
                    "patientAccessList": {
                        "pageInfo": {
                            "endCursor": None
                        },
                        "patientAccess": [
                            test_patient2
                        ]
                    }
                }
            }
        ]

        patients = get_all_patients(client=self.mock_client)

        self.assertEqual(
            [
                {
                    "id": "p1",
                    "created_at": 1630515986.9949625,
                    "name": "patient1",
                    "devices": [
                        {
                            "patient_id": "p1",
                            "name": "Percept",
                            "created_at": 1646685476.1367705,
                            "device_type_id": "dt1",
                            "disabled": False,
                            "disabled_at": None,
                            "updated_at": 1646685485.9403558,
                            "id": "d1"
                        }
                    ]
                },
                {
                    "id": "p2",
                    "created_at": 1630515986.9949625,
                    "name": "patient2",
                    "devices": [
                        {
                            "patient_id": "p2",
                            "name": "Percept",
                            "created_at": 1646685476.1367705,
                            "device_type_id": "dt1",
                            "disabled": False,
                            "disabled_at": None,
                            "updated_at": 1646685485.9403558,
                            "id": "d1"
                        }
                    ]
                }
            ],
            patients.to_list(),
        )


class TestDevice(TestCase):
    """
    Unit tests for the Device class.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.maxDiff = None
        self.mock_client = GraphClient(
            Config(
                client_key_id='test',
                client_access_key='config'
            )
        )

    def test_attributes(self):
        """
        Test attributes for an initialized Device.

        """
        test_device = Device(
            id="device-1",
            patient_id="patient-1",
            name="Percept",
            created_at=1629300943.9179766,
            device_type_id="dt1",
        )

        self.assertEqual("device-1", test_device.id)
        self.assertEqual("patient-1", test_device.patient_id)
        self.assertEqual("Percept", test_device.name)
        self.assertEqual(1629300943.9179766, test_device.created_at)
        self.assertEqual("dt1", test_device.device_type_id)

    def test_repr(self):
        """
        Test __repr__

        """
        test_device = Device(
            id="1",
            patient_id="p1",
            name="Percept",
            created_at=1629300943.9179766,
            device_type_id="dt1",
        )
        self.assertEqual(
            'Device(id="1", name="Percept", patient_id="p1")',
            repr(test_device)
        )

    def test_normalize_id(self):
        """
        Test strip id of resource prefix if it exists.

        """
        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Device.normalize_id("device-d4b1c627bd464fe0a5ed940cc8e8e485")
        )

        self.assertEqual(
            "d4b1c627bd464fe0a5ed940cc8e8e485",
            Device.normalize_id("d4b1c627bd464fe0a5ed940cc8e8e485")
        )

    def test_denormalize_id(self):
        """
        Test add resource prefix and suffix to if they don't exist.

        """
        self.assertEqual(
            "patient-d4b1c627bd4,device-d4b1c627bd46",
            Device.denormalize_id(
                "patient-d4b1c627bd4",
                "device-d4b1c627bd46"
            )
        )

        self.assertEqual(
            "patient-d4b1c627bd4,device-d4b1c627bd46",
            Device.denormalize_id(
                "patient-d4b1c627bd4",
                "d4b1c627bd46"
            )
        )

        self.assertEqual(
            "patient-d4b1c627bd4,device-d4b1c627bd46",
            Device.denormalize_id(
                "d4b1c627bd4",
                "device-d4b1c627bd46"
            )
        )

        self.assertEqual(
            "patient-d4b1c627bd4,device-d4b1c627bd46",
            Device.denormalize_id(
                "d4b1c627bd4",
                "d4b1c627bd46"
            )
        )


    def test_get_device(self):
        """
        Test get device.

        """
        test_device = Device(
            id="1",
            patient_id="p1",
            name="Percept",
            created_at=1629300943.9179766,
            device_type_id="dt1",
        )

        test_patient = Patient(
            id="p1",
            created_at=1629300943.9179766,
            name="patient1",
            devices=DeviceSet(
                [
                    test_device
                ]
            )
        )

        act_device = get_device(
            patient=test_patient,
            device_id="device-1"
        )
        self.assertEqual(test_device.to_dict(), act_device.to_dict())

    def test_get_patient_devices(self):
        """
        Test get patient devices.

        """
        test_device_1 = Device(
            id="d1",
            patient_id="p1",
            name="Percept",
            created_at=1629300943.9179766,
            device_type_id="dt1",
        )

        test_device_2 = Device(
            id="d2",
            patient_id="p1",
            name="Apple Watch",
            created_at=1629300943.9179766,
            device_type_id="dt2",
        )

        test_patient = Patient(
            id="p1",
            created_at=1629300943.9179766,
            name="patient1",
            devices=DeviceSet(
                [
                    test_device_1,
                    test_device_2
                ]
            )
        )

        act_devices = get_patient_devices(
            patient=test_patient,
        )
        self.assertEqual(
            [
                test_device_1.to_dict(),
                test_device_2.to_dict()
            ],
            act_devices.to_list()
        )

    def test_get_all_devices(self):
        """
        Test get all devices.

        """
        test_device_1 = Device(
            id="d1",
            patient_id="p1",
            name="Percept",
            created_at=1629300943.9179766,
            device_type_id="dt1",
        )

        test_patient_1 = Patient(
            id="p1",
            created_at=1629300943.9179766,
            name="patient1",
            devices=DeviceSet([test_device_1])
        )

        test_device_2 = Device(
            id="d2",
            patient_id="p1",
            name="Apple Watch",
            created_at=1629300943.9179766,
            device_type_id="dt2",
        )

        test_patient_2 = Patient(
            id="p2",
            created_at=1629300943.9179766,
            name="patient2",
            devices=DeviceSet([test_device_2])
        )

        patients = PatientSet([test_patient_1, test_patient_2])

        act_devices = get_all_devices(patients=patients)
        self.assertEqual(
            [
                test_device_1.to_dict(),
                test_device_2.to_dict()
            ],
            act_devices.to_list()
        )



"""
Common utilities and base classes.

"""
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Dict, Iterable, Iterator, List, Type, Union


class ItemBase:
    """
    Base class for representing an item.

    """
    # ID of the Item
    _id: str

    # Dictionary of attributes associated with the Item
    _attributes: dict

    def __init__(self, id: str, **attributes):
        """
        Initialize item's representational state.

        """
        self._id = id

        attributes["id"] = id
        self._attributes = attributes

    @property
    def id(self) -> str:
        """
        ID of the item.

        """
        return self._id

    def __eq__(self, right):
        """
        Compare two items for equality by ID.

        """
        return type(self) == type(right) and self._id == right.id

    def __getitem__(self, attr: str):
        """
        Get the value of any attribute in self.attributes.

        """
        return self._attributes.get(attr)

    def get(self, attr: str, default: Any = None):
        """
        Get the value of any attribute in self.attributes.

        """
        return self._attributes.get(attr, default)

    def __repr__(self):
        """
        Format the item representation.

        """
        if hasattr(self, "name"):
            return (
                f'{self.__class__.__name__}'
                f'(id="{self.id}", name="{self.name}")'
            )
        else:
            return f'{self.__class__.__name__}(id="{self.id}")'

    def to_dict(self) -> dict:
        """
        Dictionary representation of the item's attributes.

        """
        return self._attributes


class ItemSet(ABC):
    """
    Base class for representing a set of items.

    """

    # The Items in this set. Maps id to item.
    _items: Dict[str, ItemBase]

    def __init__(self, items: Iterable[ItemBase] = ()):
        """
        Initialize items set representational state.

        """
        self._items = OrderedDict()
        for item in items:
            self.add(item)

    def __iter__(self) -> Iterator[ItemBase]:
        """
        Iterate over the items of the set.

        """
        return iter(self._items.values())

    def __len__(self):
        """
        Number of items in the set.

        """
        return len(self._items)

    def __getitem__(self, id: str):
        """
        Get an item.

        """
        return self._items.get(id)

    @property
    @abstractmethod
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        pass

    def get(self, id: str) -> ItemBase:
        """
        Get an item by id.

        Raises:
            ValueError: if the set does not contain an item with the ID.

        """
        try:
            return self._items[id]
        except KeyError:
            pass

        raise ValueError(
            f"{self.__class__.__name__} does not contain "
            f"{self._item_class.__name__} with ID {id}"
        )

    def add(self, item: ItemBase):
        """
        Add a single item to the set. Must be the same type as other members
        of the collection

        """
        if type(item) is not self._item_class:
            raise TypeError(
                f'cannot add {str(item)}; must be type '
                f'{self._item_class.__name__}'
            )

        self._items[item.id] = item

    def update(self, items: Iterable[ItemBase]):
        """
        Add an iterable of item(s) to this set. All items must be the same
        class as members of this collection.

        """
        for item in items:
            if type(item) is not self._item_class:
                raise TypeError(
                    f'cannot update with {str(item)}; all items must be type '
                    f'{self._item_class.__name__}'
                )

            self._items[item.id] = item

    def remove(self, *items: Union[str, ItemBase]):
        """
        Remove item(s) from this set.

        """
        for item in items:
            if type(item) is str:
                # item is an id
                self._items.pop(item, None)
            else:
                # item is an instance of ItemBase
                self._items.pop(item.id, None)

    def ids(self) -> Iterator[str]:
        """
        Iterator over the IDs of the items in this set.

        """
        return iter(self._items.keys())

    def __repr__(self):
        """
        String representation of the set of items. Returns information on
        the number of items in the set, the type of the set, and up to 3
        items.

        """
        item_strs = []
        for i, item in enumerate(self):
            if i > 2:
                item_strs.append(
                    f'\t... (and {len(self._items.keys()) - 3} others)\n'
                )
                break

            item_strs.append(f'\t{str(item)}\n')

        items_str = ''.join(item_strs)
        return f'{type(self).__name__} {{\n{items_str}}}'

    def to_list(self) -> List[dict]:
        """
        List of all the items in the set. Each item is formatted as a
        dictionary, using the item's `to_dict()` method.

        """
        items_list = []
        for item in self._items.values():
            items_list.append(item.to_dict())

        return items_list



"""
Tests for fetching stream metadata.

"""
import copy
import json
from unittest import TestCase, mock

from runeq.config import Config
from runeq.resources.client import GraphClient, StreamClient
from runeq.resources.stream_metadata import (
    Dimension, StreamMetadata, StreamMetadataSet, StreamType,
    get_patient_stream_metadata, get_all_stream_types,
    get_stream_availability_dataframe, get_stream_dataframe,
    get_stream_metadata
)


class TestDimension(TestCase):
    """
    Unit tests for the Dimension class.

    """

    def test_attributes(self):
        """
        Test attributes for an initialized Dimension.

        """
        test_dim = Dimension(
            id="time",
            data_type="timestamp",
            quantity_name="Time",
            unit_name="Nanoseconds",
            quantity_abbrev="t",
            unitAbbrev="ns",
        )

        self.assertEqual("time", test_dim.id)
        self.assertEqual("timestamp", test_dim.data_type)
        self.assertEqual("Time", test_dim.quantity_name)
        self.assertEqual("Nanoseconds", test_dim.unit_name)


class TestStreamType(TestCase):
    """
    Unit tests for the StreamType class.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.mock_client = GraphClient(
            Config(
                client_key_id='test',
                client_access_key='config'
            )
        )

    def test_attributes(self):
        """
        Test attributes for an initialized Stream.

        """
        test_stream_type = StreamType(
            id="rotation",
            name="Rotation",
            description="Rotation rate (velocity) sampled in the time-domain.",
            dimensions=[]
        )

        self.assertEqual("rotation", test_stream_type.id)
        self.assertEqual("Rotation", test_stream_type.name)
        self.assertEqual(
            "Rotation rate (velocity) sampled in the time-domain.",
            test_stream_type.description
        )
        self.assertEqual([], test_stream_type.dimensions)

    def test_repr(self):
        """
        Test __repr__

        """
        test_stream_type = StreamType(
            id="rotation",
            name="Rotation",
            description="Rotation rate (velocity) sampled in the time-domain.",
            dimensions=[]
        )
        self.assertEqual(
            'StreamType(id="rotation", name="Rotation")',
            repr(test_stream_type)
        )

    def test_get_all_stream_types(self):
        """
        Test get all stream types.

        """
        self.mock_client.execute = mock.Mock()
        self.mock_client.execute.side_effect = [
            {
                "streamTypeList": {
                    "streamTypes": [
                        {
                            "id": "rotation",
                            "name": "Rotation",
                            "description": "Rotation rate (angular velocity)",
                            "shape": {
                                "dimensions": [
                                    {
                                        "id": "time",
                                        "data_type": "timestamp",
                                        "quantity_name": "Time",
                                        "quantity_abbrev": "t",
                                        "unit_name": "Nanoseconds",
                                        "unit_abbrev": "ns"
                                    },
                                    {
                                        "id": "rotation",
                                        "data_type": "sfloat",
                                        "quantity_name": "Angular Velocity",
                                        "quantity_abbrev": "Rotation",
                                        "unit_name": "Radians per second",
                                        "unit_abbrev": "rps"
                                    }
                                ]
                            }
                        },
                        {
                            "id": "current",
                            "name": "Current",
                            "description": "Electric current (in Amperes)",
                            "shape": {
                                "dimensions": [
                                    {
                                        "id": "time",
                                        "data_type": "timestamp",
                                        "quantity_name": "Time",
                                        "quantity_abbrev": "t",
                                        "unit_name": "Nanoseconds",
                                        "unit_abbrev": "ns"
                                    },
                                    {
                                        "id": "current",
                                        "data_type": "sfloat",
                                        "quantity_name": "Current",
                                        "quantity_abbrev": "I",
                                        "unit_name": "Amps",
                                        "unit_abbrev": "A"
                                    }
                                ]
                            }
                        }
                    ]
                }
            }
        ]

        stream_types = get_all_stream_types(client=self.mock_client)

        self.assertEqual(
            [
                {
                    "id": "rotation",
                    "name": "Rotation",
                    "description": "Rotation rate (angular velocity)",
                    "dimensions": [
                        {
                            "id": "time",
                            "data_type": "timestamp",
                            "quantity_name": "Time",
                            "quantity_abbrev": "t",
                            "unit_name": "Nanoseconds",
                            "unit_abbrev": "ns"
                        },
                        {
                            "id": "rotation",
                            "data_type": "sfloat",
                            "quantity_name": "Angular Velocity",
                            "quantity_abbrev": "Rotation",
                            "unit_name": "Radians per second",
                            "unit_abbrev": "rps"
                        }
                    ]
                },
                {
                    "id": "current",
                    "name": "Current",
                    "description": "Electric current (in Amperes)",
                    "dimensions": [
                        {
                            "id": "time",
                            "data_type": "timestamp",
                            "quantity_name": "Time",
                            "quantity_abbrev": "t",
                            "unit_name": "Nanoseconds",
                            "unit_abbrev": "ns"
                        },
                        {
                            "id": "current",
                            "data_type": "sfloat",
                            "quantity_name": "Current",
                            "quantity_abbrev": "I",
                            "unit_name": "Amps",
                            "unit_abbrev": "A"
                        }
                    ]
                }
            ],
            stream_types.to_list(),
        )


class TestStreamMetadata(TestCase):
    """
    Unit tests for the StreamMetadata class.

    """

    def setUp(self):
        """
        Set up mock graph and stream clients for testing.

        """
        self.mock_graph_client = GraphClient(
            Config(
                client_key_id='test',
                client_access_key='config'
            )
        )

        self.mock_stream_client = StreamClient(
            Config(
                client_key_id='test',
                client_access_key='config'
            )
        )

        self.maxDiff = None

    def test_attributes(self):
        """
        Test attributes for an initialized Stream.

        """
        test_stream_type = StreamType(
            id="st1",
            name="stream_type_1",
            description="stream type one",
            dimensions=[],
        )

        test_stream = StreamMetadata(
            id="s1",
            created_at=1629300943.9179766,
            algorithm="a1",
            device_id="d1",
            patient_id="p1",
            stream_type=test_stream_type,
            min_time=1648231560,
            max_time=1648234860,
            parameters={"category": "vitals"}
        )

        self.assertEqual("s1", test_stream.id)
        self.assertEqual(1629300943.9179766, test_stream.created_at)
        self.assertEqual("a1", test_stream.algorithm)
        self.assertEqual("d1", test_stream.device_id)
        self.assertEqual("p1", test_stream.patient_id)
        self.assertEqual(test_stream_type, test_stream.stream_type)
        self.assertEqual(1648231560, test_stream.min_time)
        self.assertEqual(1648234860, test_stream.max_time)
        self.assertEqual({"category": "vitals"}, test_stream.parameters)

    def test_repr(self):
        """
        Test __repr__

        """
        test_stream = StreamMetadata(
            id="s1",
            created_at=1629300943.9179766,
            algorithm="a1",
            device_id="d1",
            patient_id="p1",
            stream_type=None,
            min_time=1648231560,
            max_time=1648234860,
            parameters={"category": "vitals"}
        )
        self.assertEqual(
            'StreamMetadata(id="s1")',
            repr(test_stream)
        )

    def test_get_stream_metadata(self):
        """
        Test getting stream metadata.

        """
        self.mock_graph_client.execute = mock.Mock()
        self.mock_graph_client.execute.side_effect = [
            {
                "streamListByIds": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "streams": [
                        {
                            "id": "s1",
                            "created_at": 1655226140.508,
                            "algorithm": "a1",
                            "device_id": "patient-p1,device-d1",
                            "patient_id": "p1",
                            "streamType": {
                                "id": "duration",
                                "name": "Duration",
                                "description": "Duration over time.",
                                "shape": {
                                    "dimensions": [
                                        {
                                            "id": "time",
                                            "data_type": "timestamp",
                                            "quantity_name": "Time",
                                            "quantity_abbrev": "t",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        },
                                        {
                                            "id": "duration",
                                            "data_type": "sfloat",
                                            "quantity_name": "Duration",
                                            "quantity_abbrev": "Duration",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        }
                                    ]
                                }
                            },
                            "min_time": 1648231560,
                            "max_time": 1648234860
                        },
                        {
                            "id": "s2",
                            "created_at": 1655226140.501,
                            "algorithm": "a2",
                            "device_id": "patient-p2,device-d2",
                            "patient_id": "p2",
                            "streamType": {
                                "id": "duration",
                                "name": "Duration",
                                "description": "Duration over time.",
                                "shape": {
                                    "dimensions": [
                                        {
                                            "id": "time",
                                            "data_type": "timestamp",
                                            "quantity_name": "Time",
                                            "quantity_abbrev": "t",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        },
                                        {
                                            "id": "duration",
                                            "data_type": "sfloat",
                                            "quantity_name": "Duration",
                                            "quantity_abbrev": "Duration",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        }
                                    ]
                                }
                            },
                            "min_time": 1648231560,
                            "max_time": 1648234860
                        }
                    ]
                }
            }
        ]

        streams = get_stream_metadata(
            stream_ids=["s1", "s2"],
            client=self.mock_graph_client
        )

        self.assertEqual(
            [
                {
                    "created_at": 1655226140.508,
                    "algorithm": "a1",
                    "device_id": "d1",
                    "patient_id": "p1",
                    "stream_type": {
                        "name": "Duration",
                        "description": "Duration over time.",
                        "dimensions": [
                            {
                                "data_type": "timestamp",
                                "quantity_name": "Time",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "t",
                                "unit_abbrev": "s",
                                "id": "time"
                            },
                            {
                                "data_type": "sfloat",
                                "quantity_name": "Duration",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "Duration",
                                "unit_abbrev": "s",
                                "id": "duration"
                            }
                        ],
                        "id": "duration"
                    },
                    'parameters': {},
                    "min_time": 1648231560,
                    "max_time": 1648234860,
                    "id": "s1"
                },
                {
                    "created_at": 1655226140.501,
                    "algorithm": "a2",
                    "device_id": "d2",
                    "patient_id": "p2",
                    "stream_type": {
                        "name": "Duration",
                        "description": "Duration over time.",
                        "dimensions": [
                            {
                                "data_type": "timestamp",
                                "quantity_name": "Time",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "t",
                                "unit_abbrev": "s",
                                "id": "time"
                            },
                            {
                                "data_type": "sfloat",
                                "quantity_name": "Duration",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "Duration",
                                "unit_abbrev": "s",
                                "id": "duration"
                            }
                        ],
                        "id": "duration"
                    },
                    'parameters': {},
                    "min_time": 1648231560,
                    "max_time": 1648234860,
                    "id": "s2"
                }
            ],
            streams.to_list(),
        )

    def test_get_over_hundred_stream_metadata(self):
        """
        Test get stream metadata can query for >100 streams by batching
        by requests of size <=100 streams.

        """
        self.mock_graph_client.execute = mock.Mock()

        stream_resp = {
            "created_at": 1655226140.508,
            "algorithm": "a1",
            "device_id": "patient-p1,device-d1",
            "patient_id": "p1",
            "streamType": {
                "id": "duration",
                "name": "Duration",
                "description": "Duration over time.",
                "shape": {
                    "dimensions": [
                        {
                            "id": "time",
                            "data_type": "timestamp",
                            "quantity_name": "Time",
                            "quantity_abbrev": "t",
                            "unit_name": "Seconds",
                            "unit_abbrev": "s"
                        },
                        {
                            "id": "duration",
                            "data_type": "sfloat",
                            "quantity_name": "Duration",
                            "quantity_abbrev": "Duration",
                            "unit_name": "Seconds",
                            "unit_abbrev": "s"
                        }
                    ]
                }
            },
            "min_time": 1648231560,
            "max_time": 1648234860
        }

        first_hundred_streams = []
        for i in range(100):
            resp = copy.deepcopy(stream_resp)
            resp["id"] = str(i)
            first_hundred_streams.append(resp)

        next_fifty_streams = []
        for i in range(100, 150):
            resp = copy.deepcopy(stream_resp)
            resp["id"] = str(i)
            next_fifty_streams.append(resp)

        self.mock_graph_client.execute.side_effect = [
            {
                "streamListByIds": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "streams": first_hundred_streams
                }
            },
            {
                "streamListByIds": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "streams": next_fifty_streams
                }
            },
        ]

        streams = get_stream_metadata(
            stream_ids=[str(i) for i in range(150)],
            client=self.mock_graph_client
        )

        self.assertEqual(150, len(streams.to_list()))

    @mock.patch("runeq.resources.stream_metadata.get_patient")
    def test_get_patient_stream_metadata_no_access(self, get_patient):
        """
        Test get_patient_stream_metadata fails if the user doesn't have access
        to the patient ID or if the patient doesn't exist.
        """
        get_patient.side_effect = Exception("NotFoundError")

        with self.assertRaises(Exception) as context:
            get_patient_stream_metadata(
                patient_id='foo',
                client=self.mock_graph_client
            )

        self.assertTrue('NotFoundError' in str(context.exception))

    @mock.patch("runeq.resources.stream_metadata.get_patient")
    def test_get_patient_streams_basic(self, _):
        """
        Test filtering streams by all parameters.

        """
        self.mock_graph_client.execute = mock.Mock()
        self.mock_graph_client.execute.side_effect = [
            {
                "streamList": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "streams": [
                        {
                            "id": "s1",
                            "created_at": 1655226140.508,
                            "algorithm": "a1",
                            "device_id": "patient-p1,device-d1",
                            "patient_id": "p1",
                            "streamType": {
                                "id": "duration",
                                "name": "Duration",
                                "description": "Duration over time.",
                                "shape": {
                                    "dimensions": [
                                        {
                                            "id": "time",
                                            "data_type": "timestamp",
                                            "quantity_name": "Time",
                                            "quantity_abbrev": "t",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        },
                                        {
                                            "id": "duration",
                                            "data_type": "sfloat",
                                            "quantity_name": "Duration",
                                            "quantity_abbrev": "Duration",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        }
                                    ]
                                }
                            },
                            "parameters": [
                                {
                                    "key": "category",
                                    "value": "motion"
                                },
                                {
                                    "key": "measurement",
                                    "value": "walking"
                                }
                            ],
                            "min_time": 1648231560,
                            "max_time": 1648234860
                        },
                        {
                            "id": "s2",
                            "created_at": 1655226140.501,
                            "algorithm": "a2",
                            "device_id": "patient-p2,device-d2",
                            "patient_id": "p1",
                            "streamType": {
                                "id": "duration",
                                "name": "Duration",
                                "description": "Duration over time.",
                                "shape": {
                                    "dimensions": [
                                        {
                                            "id": "time",
                                            "data_type": "timestamp",
                                            "quantity_name": "Time",
                                            "quantity_abbrev": "t",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        },
                                        {
                                            "id": "duration",
                                            "data_type": "sfloat",
                                            "quantity_name": "Duration",
                                            "quantity_abbrev": "Duration",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        }
                                    ]
                                }
                            },
                            "min_time": 1648231560,
                            "max_time": 1648234860
                        }
                    ]
                }
            }
        ]

        streams = get_patient_stream_metadata(
            patient_id="p1",
            client=self.mock_graph_client
        )

        self.assertEqual(
            [
                {
                    "created_at": 1655226140.508,
                    "algorithm": "a1",
                    "device_id": "d1",
                    "patient_id": "p1",
                    "stream_type": {
                        "name": "Duration",
                        "description": "Duration over time.",
                        "dimensions": [
                            {
                                "data_type": "timestamp",
                                "quantity_name": "Time",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "t",
                                "unit_abbrev": "s",
                                "id": "time"
                            },
                            {
                                "data_type": "sfloat",
                                "quantity_name": "Duration",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "Duration",
                                "unit_abbrev": "s",
                                "id": "duration"
                            }
                        ],
                        "id": "duration"
                    },
                    "category": "motion",
                    "measurement": "walking",
                    'parameters': {
                        'category': 'motion',
                        'measurement': 'walking'
                    },
                    "min_time": 1648231560,
                    "max_time": 1648234860,
                    "id": "s1"
                },
                {
                    "created_at": 1655226140.501,
                    "algorithm": "a2",
                    "device_id": "d2",
                    "patient_id": "p1",
                    "stream_type": {
                        "name": "Duration",
                        "description": "Duration over time.",
                        "dimensions": [
                            {
                                "data_type": "timestamp",
                                "quantity_name": "Time",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "t",
                                "unit_abbrev": "s",
                                "id": "time"
                            },
                            {
                                "data_type": "sfloat",
                                "quantity_name": "Duration",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "Duration",
                                "unit_abbrev": "s",
                                "id": "duration"
                            }
                        ],
                        "id": "duration"
                    },
                    'parameters': {},
                    "min_time": 1648231560,
                    "max_time": 1648234860,
                    "id": "s2"
                }
            ],
            streams.to_list(),
        )

    @mock.patch("runeq.resources.stream_metadata.get_patient")
    def test_get_patient_streams_paginated(self, _):
        """
        Test filtering streams by all parameters, where the user has to
        page through streams.

        """
        self.mock_graph_client.execute = mock.Mock()
        self.mock_graph_client.execute.side_effect = [
            {
                "streamList": {
                    "pageInfo": {
                        "endCursor": "test_check_next"
                    },
                    "streams": [
                        {
                            "id": "s1",
                            "created_at": 1655226140.508,
                            "algorithm": "a1",
                            "device_id": "patient-p1,device-d1",
                            "patient_id": "p1",
                            "streamType": {
                                "id": "duration",
                                "name": "Duration",
                                "description": "Duration over time.",
                                "shape": {
                                    "dimensions": [
                                        {
                                            "id": "time",
                                            "data_type": "timestamp",
                                            "quantity_name": "Time",
                                            "quantity_abbrev": "t",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        },
                                        {
                                            "id": "duration",
                                            "data_type": "sfloat",
                                            "quantity_name": "Duration",
                                            "quantity_abbrev": "Duration",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        }
                                    ]
                                }
                            },
                            "min_time": 1648231560,
                            "max_time": 1648234860
                        },
                    ]
                }
            },
            {
                "streamList": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "streams": [
                        {
                            "id": "s2",
                            "created_at": 1655226140.501,
                            "algorithm": "a2",
                            "device_id": "patient-p2,device-d2",
                            "patient_id": "p1",
                            "streamType": {
                                "id": "duration",
                                "name": "Duration",
                                "description": "Duration over time.",
                                "shape": {
                                    "dimensions": [
                                        {
                                            "id": "time",
                                            "data_type": "timestamp",
                                            "quantity_name": "Time",
                                            "quantity_abbrev": "t",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        },
                                        {
                                            "id": "duration",
                                            "data_type": "sfloat",
                                            "quantity_name": "Duration",
                                            "quantity_abbrev": "Duration",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        }
                                    ]
                                }
                            },
                            "min_time": 1648231560,
                            "max_time": 1648234860
                        }
                    ]
                }
            }
        ]

        streams = get_patient_stream_metadata(
            patient_id="p1",
            client=self.mock_graph_client
        )

        self.assertEqual(
            [
                {
                    "created_at": 1655226140.508,
                    "algorithm": "a1",
                    "device_id": "d1",
                    "patient_id": "p1",
                    "stream_type": {
                        "name": "Duration",
                        "description": "Duration over time.",
                        "dimensions": [
                            {
                                "data_type": "timestamp",
                                "quantity_name": "Time",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "t",
                                "unit_abbrev": "s",
                                "id": "time"
                            },
                            {
                                "data_type": "sfloat",
                                "quantity_name": "Duration",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "Duration",
                                "unit_abbrev": "s",
                                "id": "duration"
                            }
                        ],
                        "id": "duration"
                    },
                    'parameters': {},
                    "min_time": 1648231560,
                    "max_time": 1648234860,
                    "id": "s1"
                },
                {
                    "created_at": 1655226140.501,
                    "algorithm": "a2",
                    "device_id": "d2",
                    "patient_id": "p1",
                    "stream_type": {
                        "name": "Duration",
                        "description": "Duration over time.",
                        "dimensions": [
                            {
                                "data_type": "timestamp",
                                "quantity_name": "Time",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "t",
                                "unit_abbrev": "s",
                                "id": "time"
                            },
                            {
                                "data_type": "sfloat",
                                "quantity_name": "Duration",
                                "unit_name": "Seconds",
                                "quantity_abbrev": "Duration",
                                "unit_abbrev": "s",
                                "id": "duration"
                            }
                        ],
                        "id": "duration"
                    },
                    'parameters': {},
                    "min_time": 1648231560,
                    "max_time": 1648234860,
                    "id": "s2"
                }
            ],
            streams.to_list(),
        )

    def test_stream_set_filter(self):
        """
        Test filtering streams within a stream set.

        """
        stream_set = StreamMetadataSet(
            [
                StreamMetadata(
                    id="stream1",
                    created_at=123,
                    algorithm="alg1",
                    device_id="d1",
                    patient_id="p1",
                    stream_type=StreamType(
                        id="acceleration",
                        name="Acceleration",
                        description="Acceleration rate",
                        dimensions=[]
                    ),
                    min_time=10,
                    max_time=100,
                    parameters={}
                ),
                StreamMetadata(
                    id="stream2",
                    created_at=123,
                    algorithm="alg1",
                    device_id="d2",
                    patient_id="p1",
                    stream_type=StreamType(
                        id="acceleration",
                        name="Acceleration",
                        description="Acceleration rate",
                        dimensions=[]
                    ),
                    min_time=10,
                    max_time=100,
                    parameters={}
                ),
                StreamMetadata(
                    id="stream3",
                    created_at=123,
                    algorithm="alg1",
                    device_id="d1",
                    patient_id="p1",
                    stream_type=StreamType(
                        id="motion",
                        name="Motion",
                        description="Motion movement",
                        dimensions=[]
                    ),
                    min_time=10,
                    max_time=100,
                    parameters={}
                ),
            ]
        )

        device_streams = stream_set.filter(
            patient_id="p1",
            device_id="d1"
        )
        self.assertEqual(2, len(device_streams))
        self.assertIsNotNone(device_streams.get("stream1"))
        self.assertIsNotNone(device_streams.get("stream3"))

        motion_streams = device_streams.filter(
            stream_type_id="motion"
        )
        self.assertEqual(1, len(motion_streams))
        self.assertIsNotNone(motion_streams.get("stream3"))

        stream1_streams = stream_set.filter(
            filter_function=lambda stream: stream.id == "stream1"
        )
        self.assertEqual(1, len(stream1_streams))
        self.assertIsNotNone(stream1_streams.get("stream1"))


    def test_get_stream_dataframe(self):
        """
        Test get stream as dataframe with stream metadata and data.

        """
        self.mock_stream_client.get_data = mock.Mock()
        self.mock_stream_client.get_data.return_value = iter(
            [
                """time,acceleration,measurement_duration_ns
2022-07-28T14:26:45.167568Z,0.020525138825178146,20000000
2022-07-28T14:26:45.361596Z,0.020834974944591522,20000000
2022-07-28T14:26:45.361796Z,0.021182861179113388,20000000
2022-07-28T14:26:45.3618588Z,0.022172993049025536,20000000
2022-07-28T14:26:45.3620749Z,0.02356025017797947,20000000
2022-07-28T14:26:45.362149Z,0.024860087782144547,20000000
2022-07-28T14:26:45.36221Z,0.026072751730680466,20000000"""
            ]
        )

        self.mock_graph_client.execute = mock.Mock()
        self.mock_graph_client.execute.side_effect = [
            {
                "streamListByIds": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "streams": [
                        {
                            "id": "s1",
                            "created_at": 1655226140.508,
                            "algorithm": "a1",
                            "device_id": "patient-p1,device-d1",
                            "patient_id": "p1",
                            "streamType": {
                                "id": "duration",
                                "name": "Duration",
                                "description": "Duration over time.",
                                "shape": {
                                    "dimensions": [
                                        {
                                            "id": "time",
                                            "data_type": "timestamp",
                                            "quantity_name": "Time",
                                            "quantity_abbrev": "t",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        },
                                        {
                                            "id": "duration",
                                            "data_type": "sfloat",
                                            "quantity_name": "Duration",
                                            "quantity_abbrev": "Duration",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        }
                                    ]
                                }
                            },
                            "min_time": 1648231560,
                            "max_time": 1648234860
                        }
                    ]
                }
            }
        ]

        stream_df = get_stream_dataframe(
            stream_ids="s1",
            stream_client=self.mock_stream_client,
            graph_client=self.mock_graph_client
        )

        self.assertEqual(
            {
                "time": {
                    "0": "2022-07-28T14:26:45.167568Z",
                    "1": "2022-07-28T14:26:45.361596Z",
                    "2": "2022-07-28T14:26:45.361796Z",
                    "3": "2022-07-28T14:26:45.3618588Z",
                    "4": "2022-07-28T14:26:45.3620749Z",
                    "5": "2022-07-28T14:26:45.362149Z",
                    "6": "2022-07-28T14:26:45.36221Z"
                },
                "acceleration": {
                    "0": 0.0205251388,
                    "1": 0.0208349749,
                    "2": 0.0211828612,
                    "3": 0.022172993,
                    "4": 0.0235602502,
                    "5": 0.0248600878,
                    "6": 0.0260727517
                },
                "measurement_duration_ns": {
                    "0": 20000000,
                    "1": 20000000,
                    "2": 20000000,
                    "3": 20000000,
                    "4": 20000000,
                    "5": 20000000,
                    "6": 20000000
                },
                "algorithm": {
                    "0": "a1",
                    "1": "a1",
                    "2": "a1",
                    "3": "a1",
                    "4": "a1",
                    "5": "a1",
                    "6": "a1"
                },
                "device_id": {
                    "0": "d1",
                    "1": "d1",
                    "2": "d1",
                    "3": "d1",
                    "4": "d1",
                    "5": "d1",
                    "6": "d1"
                },
                "patient_id": {
                    "0": "p1",
                    "1": "p1",
                    "2": "p1",
                    "3": "p1",
                    "4": "p1",
                    "5": "p1",
                    "6": "p1"
                },
                "stream_type_id": {
                    "0": "duration",
                    "1": "duration",
                    "2": "duration",
                    "3": "duration",
                    "4": "duration",
                    "5": "duration",
                    "6": "duration",
                },
                "stream_id": {
                    "0": "s1",
                    "1": "s1",
                    "2": "s1",
                    "3": "s1",
                    "4": "s1",
                    "5": "s1",
                    "6": "s1"
                }
            },
            json.loads(stream_df.to_json())
        )

    def test_get_multiple_stream_dataframe(self):
        """
        Test get multiple streams as a dataframe

        """
        self.mock_stream_client.get_data = mock.Mock()
        self.mock_stream_client.get_data.side_effect = [
            iter(["""time,acceleration,measurement_duration_ns
2022-07-28T14:26:45.167568Z,0.014469802379608154,20000000
2022-07-28T14:26:45.361596Z,0.03278458118438721,20000000
2022-07-28T14:26:45.361796Z,0.03711885213851929,20000000
2022-07-28T14:26:45.3618588Z,0.02531599998474121,20000000
2022-07-28T14:26:45.3620749Z,0.03168576955795288,20000000"""]),
            iter(["""time,acceleration,measurement_duration_ns
2022-07-28T14:26:45.167568Z,0.020525138825178146,20000000
2022-07-28T14:26:45.361596Z,0.020834974944591522,20000000
2022-07-28T14:26:45.361796Z,0.021182861179113388,20000000
2022-07-28T14:26:45.3618588Z,0.022172993049025536,20000000
2022-07-28T14:26:45.3620749Z,0.02356025017797947,20000000"""])
        ]

        self.mock_graph_client.execute = mock.Mock()
        self.mock_graph_client.execute.side_effect = [
            {
                "streamListByIds": {
                    "pageInfo": {
                        "endCursor": "None"
                    },
                    "streams": [
                        {
                            "id": "s1",
                            "created_at": 1659018576.73,
                            "algorithm": "a1",
                            "device_id": "patient-p1,device-d1",
                            "patient_id": "p1",
                            "streamType": {
                                "id": "acceleration",
                                "name": "Acceleration",
                                "description": "Acceleration sampled in time.",
                                "shape": {
                                    "dimensions": [
                                        {
                                            "id": "time",
                                            "data_type": "timestamp",
                                            "quantity_name": "Time",
                                            "quantity_abbrev": "t",
                                            "unit_name": "Nanoseconds",
                                            "unit_abbrev": "ns"
                                        },
                                        {
                                            "id": "acceleration",
                                            "data_type": "sfloat",
                                            "quantity_name": "Acceleration",
                                            "quantity_abbrev": "Accel",
                                            "unit_name": "Gs",
                                            "unit_abbrev": "G"
                                        }
                                    ]
                                }
                            },
                            "parameters": [
                                {
                                    "key": "axis",
                                    "value": "z"
                                },
                                {
                                    "key": "category",
                                    "value": "motion"
                                },
                                {
                                    "key": "measurement",
                                    "value": "user"
                                }
                            ],
                            "min_time": 1659018405.167568,
                            "max_time": 1659027683.492028
                        },
                        {
                            "id": "s2",
                            "created_at": 1659018576.448,
                            "algorithm": "a1",
                            "device_id": "patient-p1,device-d1",
                            "patient_id": "p1",
                            "streamType": {
                                "id": "acceleration",
                                "name": "Acceleration",
                                "description": "Acceleration sampled in time.",
                                "shape": {
                                    "dimensions": [
                                        {
                                            "id": "time",
                                            "data_type": "timestamp",
                                            "quantity_name": "Time",
                                            "quantity_abbrev": "t",
                                            "unit_name": "Nanoseconds",
                                            "unit_abbrev": "ns"
                                        },
                                        {
                                            "id": "acceleration",
                                            "data_type": "sfloat",
                                            "quantity_name": "Acceleration",
                                            "quantity_abbrev": "Accel",
                                            "unit_name": "Gs",
                                            "unit_abbrev": "G"
                                        }
                                    ]
                                }
                            },
                            "parameters": [
                                {
                                    "key": "axis",
                                    "value": "x"
                                },
                                {
                                    "key": "category",
                                    "value": "motion"
                                },
                                {
                                    "key": "measurement",
                                    "value": "gravity"
                                }
                            ],
                            "min_time": 1659018405.167568,
                            "max_time": 1659027683.492028
                        }
                    ]
                }
            }
        ]

        stream_df = get_stream_dataframe(
            stream_ids=["s1", "s2"],
            stream_client=self.mock_stream_client,
            graph_client=self.mock_graph_client
        )

        self.assertEqual(
            {
                "time": {
                    "0": "2022-07-28T14:26:45.167568Z",
                    "1": "2022-07-28T14:26:45.361596Z",
                    "2": "2022-07-28T14:26:45.361796Z",
                    "3": "2022-07-28T14:26:45.3618588Z",
                    "4": "2022-07-28T14:26:45.3620749Z",
                    "5": "2022-07-28T14:26:45.167568Z",
                    "6": "2022-07-28T14:26:45.361596Z",
                    "7": "2022-07-28T14:26:45.361796Z",
                    "8": "2022-07-28T14:26:45.3618588Z",
                    "9": "2022-07-28T14:26:45.3620749Z"
                },
                "acceleration": {
                    "0": 0.0144698024,
                    "1": 0.0327845812,
                    "2": 0.0371188521,
                    "3": 0.025316,
                    "4": 0.0316857696,
                    "5": 0.0205251388,
                    "6": 0.0208349749,
                    "7": 0.0211828612,
                    "8": 0.0221729930,
                    "9": 0.0235602502
                },
                "measurement_duration_ns": {
                    "0": 20000000,
                    "1": 20000000,
                    "2": 20000000,
                    "3": 20000000,
                    "4": 20000000,
                    "5": 20000000,
                    "6": 20000000,
                    "7": 20000000,
                    "8": 20000000,
                    "9": 20000000
                },
                "algorithm": {
                    "0": "a1",
                    "1": "a1",
                    "2": "a1",
                    "3": "a1",
                    "4": "a1",
                    "5": "a1",
                    "6": "a1",
                    "7": "a1",
                    "8": "a1",
                    "9": "a1"
                },
                "device_id": {
                    "0": "d1",
                    "1": "d1",
                    "2": "d1",
                    "3": "d1",
                    "4": "d1",
                    "5": "d1",
                    "6": "d1",
                    "7": "d1",
                    "8": "d1",
                    "9": "d1"
                },
                "patient_id": {
                    "0": "p1",
                    "1": "p1",
                    "2": "p1",
                    "3": "p1",
                    "4": "p1",
                    "5": "p1",
                    "6": "p1",
                    "7": "p1",
                    "8": "p1",
                    "9": "p1"
                },
                "axis": {
                    "0": "z",
                    "1": "z",
                    "2": "z",
                    "3": "z",
                    "4": "z",
                    "5": "x",
                    "6": "x",
                    "7": "x",
                    "8": "x",
                    "9": "x"
                },
                "category": {
                    "0": "motion",
                    "1": "motion",
                    "2": "motion",
                    "3": "motion",
                    "4": "motion",
                    "5": "motion",
                    "6": "motion",
                    "7": "motion",
                    "8": "motion",
                    "9": "motion"
                },
                "measurement": {
                    "0": "user",
                    "1": "user",
                    "2": "user",
                    "3": "user",
                    "4": "user",
                    "5": "gravity",
                    "6": "gravity",
                    "7": "gravity",
                    "8": "gravity",
                    "9": "gravity"
                },
                "stream_type_id": {
                    "0": "acceleration",
                    "1": "acceleration",
                    "2": "acceleration",
                    "3": "acceleration",
                    "4": "acceleration",
                    "5": "acceleration",
                    "6": "acceleration",
                    "7": "acceleration",
                    "8": "acceleration",
                    "9": "acceleration"
                },
                "stream_id": {
                    "0": "s1",
                    "1": "s1",
                    "2": "s1",
                    "3": "s1",
                    "4": "s1",
                    "5": "s2",
                    "6": "s2",
                    "7": "s2",
                    "8": "s2",
                    "9": "s2"
                }
            },
            json.loads(stream_df.to_json())
        )

    def test_get_stream_availability_dataframe(self):
        """
        Test get stream as dataframe with stream metadata and availability.

        """
        self.mock_stream_client.get_data = mock.Mock()
        self.mock_stream_client.get_data.return_value = iter(
            [
                """time,availability
2022-07-28T14:26:45.167568Z,1
2022-07-28T14:26:45.361596Z,1
2022-07-28T14:26:45.361796Z,0
2022-07-28T14:26:45.3618588Z,0
2022-07-28T14:26:45.3620749Z,0
2022-07-28T14:26:45.362149Z,1
2022-07-28T14:26:45.36221Z,1
"""
            ]
        )

        self.mock_graph_client.execute = mock.Mock()
        self.mock_graph_client.execute.side_effect = [
            {
                "streamListByIds": {
                    "pageInfo": {
                        "endCursor": None
                    },
                    "streams": [
                        {
                            "id": "s1",
                            "created_at": 1655226140.508,
                            "algorithm": "a1",
                            "device_id": "patient-p1,device-d1",
                            "patient_id": "p1",
                            "streamType": {
                                "id": "duration",
                                "name": "Duration",
                                "description": "Duration over time.",
                                "shape": {
                                    "dimensions": [
                                        {
                                            "id": "time",
                                            "data_type": "timestamp",
                                            "quantity_name": "Time",
                                            "quantity_abbrev": "t",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        },
                                        {
                                            "id": "duration",
                                            "data_type": "sfloat",
                                            "quantity_name": "Duration",
                                            "quantity_abbrev": "Duration",
                                            "unit_name": "Seconds",
                                            "unit_abbrev": "s"
                                        }
                                    ]
                                }
                            },
                            "min_time": 1648231560,
                            "max_time": 1648234860
                        }
                    ]
                }
            }
        ]

        stream_df = get_stream_availability_dataframe(
            stream_ids="s1",
            start_time=123,
            end_time=345,
            resolution=300,
            stream_client=self.mock_stream_client,
            graph_client=self.mock_graph_client
        )

        self.assertEqual(
            {
                "time": {
                    "0": "2022-07-28T14:26:45.167568Z",
                    "1": "2022-07-28T14:26:45.361596Z",
                    "2": "2022-07-28T14:26:45.361796Z",
                    "3": "2022-07-28T14:26:45.3618588Z",
                    "4": "2022-07-28T14:26:45.3620749Z",
                    "5": "2022-07-28T14:26:45.362149Z",
                    "6": "2022-07-28T14:26:45.36221Z"
                },
                "availability": {
                    "0": 1,
                    "1": 1,
                    "2": 0,
                    "3": 0,
                    "4": 0,
                    "5": 1,
                    "6": 1
                },
                "algorithm": {
                    "0": "a1",
                    "1": "a1",
                    "2": "a1",
                    "3": "a1",
                    "4": "a1",
                    "5": "a1",
                    "6": "a1"
                },
                "device_id": {
                    "0": "d1",
                    "1": "d1",
                    "2": "d1",
                    "3": "d1",
                    "4": "d1",
                    "5": "d1",
                    "6": "d1"
                },
                "patient_id": {
                    "0": "p1",
                    "1": "p1",
                    "2": "p1",
                    "3": "p1",
                    "4": "p1",
                    "5": "p1",
                    "6": "p1"
                },
                "stream_type_id": {
                    "0": "duration",
                    "1": "duration",
                    "2": "duration",
                    "3": "duration",
                    "4": "duration",
                    "5": "duration",
                    "6": "duration",
                },
                "stream_id": {
                    "0": "s1",
                    "1": "s1",
                    "2": "s1",
                    "3": "s1",
                    "4": "s1",
                    "5": "s1",
                    "6": "s1"
                }
            },
            json.loads(stream_df.to_json())
        )

    def test_get_batch_stream_availability_dataframe(self):
        """
        Test get stream availability as dataframe with multiple stream results
        in a simplified dataframe without metadata.

        """
        self.mock_stream_client.get_data = mock.Mock()
        self.mock_stream_client.get_data.return_value = iter(
            [
                """time,availability
2022-07-28T14:26:45.167568Z,1
2022-07-28T14:26:45.361596Z,1
2022-07-28T14:26:45.361796Z,0
2022-07-28T14:26:45.3618588Z,0
2022-07-28T14:26:45.3620749Z,0
2022-07-28T14:26:45.362149Z,1
2022-07-28T14:26:45.36221Z,1
"""
            ]
        )

        stream_df = get_stream_availability_dataframe(
            stream_ids=["s1", "s2"],
            start_time=123,
            end_time=345,
            resolution=300,
            batch_operation="all",
            stream_client=self.mock_stream_client
        )

        self.assertEqual(
            {
                "time": {
                    "0": "2022-07-28T14:26:45.167568Z",
                    "1": "2022-07-28T14:26:45.361596Z",
                    "2": "2022-07-28T14:26:45.361796Z",
                    "3": "2022-07-28T14:26:45.3618588Z",
                    "4": "2022-07-28T14:26:45.3620749Z",
                    "5": "2022-07-28T14:26:45.362149Z",
                    "6": "2022-07-28T14:26:45.36221Z"
                },
                "availability": {
                    "0": 1,
                    "1": 1,
                    "2": 0,
                    "3": 0,
                    "4": 0,
                    "5": 1,
                    "6": 1
                },
            },
            json.loads(stream_df.to_json())
        )



import os
from unittest import TestCase

from runeq import Config

TEST_ROOT = os.path.dirname(os.path.realpath(__file__))


class TestConfig(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.access_token_good_config = (
            f'{TEST_ROOT}/data/access_token_good_config.yaml'
        )
        cls.access_token_bad_config = (
            f'{TEST_ROOT}/data/access_token_bad_config.yaml'
        )
        cls.client_keys_good_config = (
            f'{TEST_ROOT}/data/client_keys_good_config.yaml'
        )
        cls.client_keys_bad_config = (
            f'{TEST_ROOT}/data/client_keys_bad_config.yaml'
        )

    def test_init_file(self):
        """
        Test initializing using a yaml file
        """
        # Test access tokens init
        cfg = Config(self.access_token_good_config)
        expected = {
            'X-Rune-User-Access-Token-Id': 'foo',
            'X-Rune-User-Access-Token-Secret': 'bar',
        }
        self.assertEqual(expected, cfg.auth_headers)
        self.assertEqual(expected, cfg.access_token_auth_headers)

        with self.assertRaises(ValueError):
            _ = cfg.jwt_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.client_auth_headers

        # Test client keys init
        cfg = Config(self.client_keys_good_config)

        expected = {
            'X-Rune-Client-Key-ID': 'abc',
            'X-Rune-Client-Access-Key': 'def',
        }
        self.assertEqual(expected, cfg.auth_headers)
        self.assertEqual(expected, cfg.client_auth_headers)
        with self.assertRaises(ValueError):
            _ = cfg.jwt_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.access_token_auth_headers

        self.assertEqual('https://foo.runelabs.io', cfg.stream_url)

    def test_init_file_invalid(self):
        """
        Test initializing with an invalid YAML file
        """
        # Test invalid access token file
        with self.assertRaises(ValueError):
            Config(self.access_token_bad_config)

        # Test invalid client keys file
        with self.assertRaises(ValueError):
            Config(self.client_keys_bad_config)

    def test_init_kwargs(self):
        """
        Test initializing with valid kwargs
        """
        # access token
        cfg = Config(
            access_token_id='foo',
            access_token_secret='bar'
        )
        access_token_headers = {
            'X-Rune-User-Access-Token-Id': 'foo',
            'X-Rune-User-Access-Token-Secret': 'bar',
        }
        self.assertEqual(
            access_token_headers,
            cfg.auth_headers
        )
        self.assertEqual(
            access_token_headers,
            cfg.access_token_auth_headers
        )
        with self.assertRaises(ValueError):
            _ = cfg.client_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.jwt_auth_headers

        # client keys
        cfg = Config(
            client_key_id='abc',
            client_access_key='123',
        )
        client_headers = {
            'X-Rune-Client-Key-ID': 'abc',
            'X-Rune-Client-Access-Key': '123',
        }
        self.assertEqual(client_headers, cfg.auth_headers)
        self.assertEqual(client_headers, cfg.client_auth_headers)
        with self.assertRaises(ValueError):
            _ = cfg.jwt_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.access_token_auth_headers

        # jwt
        cfg = Config(jwt='abc123')
        jwt_headers = {
            'X-Rune-User-Access-Token': 'abc123'
        }
        self.assertEqual(jwt_headers, cfg.auth_headers)
        self.assertEqual(jwt_headers, cfg.jwt_auth_headers)
        with self.assertRaises(ValueError):
            _ = cfg.client_auth_headers

        with self.assertRaises(ValueError):
            _ = cfg.access_token_auth_headers

    def test_init_kwargs_invalid(self):
        """
        Test initializing with invalid kwargs.

        """
        with self.assertRaises(ValueError):
            Config(client_key_id='abc')

        with self.assertRaises(ValueError):
            Config(client_access_key='123')

        with self.assertRaises(ValueError):
            Config(client_access_key='123', jwt='abc123')

        with self.assertRaises(ValueError):
            Config(access_token_id='foo')

        with self.assertRaises(ValueError):
            Config(
                access_token_id='foo',
                client_access_key='123'
            )

        with self.assertRaises(ValueError):
            Config(
                access_token_id='foo',
                client_access_key='123',
                jwt='zee'
            )

    def test_init_invalid(self):
        """
        Cannot initialize with a filename and kwargs
        """
        with self.assertRaises(TypeError):
            Config(
                self.client_keys_good_config,
                stream_url='https://bar.runelabs.io'
            )



"""
Clients for Rune's GraphQL API and V2 Stream API.

"""
from functools import wraps
import time
from typing import Dict, Iterator, Union
import urllib.parse

from gql import Client as GQLClient
from gql import gql
from gql.transport.requests import RequestsHTTPTransport
import requests

from runeq import Config, errors

# Error when a client is not initialized
INITIALIZATION_ERROR = errors.InitializationError(
    'runeq must be initialized by calling'
    '`initialize` before this function can be used'
)

# Rune GraphQL Client to query stream metadata.
_graph_client = None

# Rune Stream API Client to query stream data.
_stream_client = None


def _retry(exceptions, max_attempts=3, max_sleep_secs=0):
    """
    Returns an exponential retry decorator.

    """
    if isinstance(exceptions, type):
        exceptions = (exceptions,)
    else:
        exceptions = tuple(exceptions)

    def inner_func(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for attempt in range(max_attempts):
                try:
                    return func(*args, **kwargs)
                except exceptions:
                    if attempt == max_attempts - 1:
                        raise

                wait_secs = 2 ** attempt
                if max_sleep_secs > 0:
                    wait_secs = min(wait_secs, max_sleep_secs)

                time.sleep(wait_secs)

        return wrapper

    return inner_func


class GraphClient:
    """
    Rune GraphQL Client to query stream metadata.

    """

    # Configuration details for the graph client.
    config: Config = None

    # The GraphQL client.
    _gql_client: GQLClient

    def __init__(self, config: Config):
        """
        Initialize the Graph API Client.

        """
        self.config = config
        transport = RequestsHTTPTransport(
            # NOTE: retries are managed by the requests.HTTPAdapter, which
            # doesn't retry failed connections
            retries=3,
            url=f"{config.graph_url}/graphql",
            use_json=True,
            headers={
                "Content-Type": "application/json",
                **config.auth_headers
            },
        )
        self._gql_client = GQLClient(
            transport=transport,
        )

    @_retry(requests.exceptions.ConnectionError)
    def execute(self, statement: str, **variables) -> Dict:
        """
        Execute a GraphQL query against the API.

        """
        return self._gql_client.execute(
            gql(statement),
            variable_values=variables
        )


class StreamClient:
    """
    Client to query the V2 Stream API.

    """

    # Pagination token to get the next page of results.
    HEADER_NEXT_PAGE = "X-Rune-Next-Page-Token"

    # Configuration details for the stream client.
    config: Config = None

    def __init__(self, config: Config):
        """
        Initialize the Stream API Client.

        """
        self.config = config

    def get_data(self, path: str, **params) -> Iterator[Union[str, dict]]:
        """
        Makes request(s) to an endpoint of the V2 Stream API. Iterates over
        responses, following pagination headers until all data has been
        fetched.

        Args:
            path: Path for an endpoint of the V2 Stream API.
            **params: Query parameters. If the format parameter is "json",
                responses are parsed as JSON. Otherwise, the response is
                returned as text.

        Returns:
            Iterator over the API responses. If the format parameter is "json",
            each value is a dictionary. Otherwise, each value is a
            CSV-formatted string (the default for the V2 API).

        Raises:
            errors.APIError

        """
        if not path.startswith("/v2"):
            raise ValueError("path must begin with /v2")

        url = urllib.parse.urljoin(self.config.stream_url, path)

        # V2 endpoints return CSV-formatted responses by default
        return_json = params.get("format") == "json"

        while True:
            r = requests.get(
                url,
                headers=self.config.auth_headers,
                params=params
            )

            self._check_response(r)
            r.raise_for_status()

            if return_json:
                yield r.json()
            else:
                yield r.text

            if self.HEADER_NEXT_PAGE not in r.headers:
                return

            params["page_token"] = r.headers[self.HEADER_NEXT_PAGE]

    def _check_response(self, r: requests.Response) -> None:
        """
        Raise an exception if the request was not successful.

        """
        if r.ok:
            return

        # When possible, the API returns details about what went wrong in
        # the json body of the response. Incorporate that detail into the error
        # raised.
        try:
            data = r.json()
        except Exception:
            r.raise_for_status()
            return

        if "error" in data:
            raise errors.APIError(r.status_code, data["error"])


def initialize(*args, **kwargs):
    """
    Initializes the library with specified configuration options. Sets global
    clients for requests to the GraphQL API and the V2 Stream API.

    Parameters
    ----------
    *args
        Accepts at most 1; a filename. If provided, values will be
        loaded from the file. It is invalid to provide both a filename
        **and** keyword arguments.
    **kwargs
        Initialize client with keyword arguments.
        If using client keys, specify the client_key_id & client_access_key.
        If using access tokens, specify access_token_id & access_token_secret.

    Examples
    --------
    There are several valid ways to use this function:

    >>> initialize()
    # Load the default config file (~/.rune/config)

    >>> initialize('./example_config.yaml')
    # Load values from a YAML file at a specified path

    >>> initialize(access_token_id='foo', access_token_secret='bar')
    >>> initialize(client_key_id='foo', client_access_key='bar')
    # Set configuration values using keyword arguments (instead
    # of a file). This can be used with any valid combination of
    # config options (e.g. with an access token OR a client key).

    """
    config = Config(*args, **kwargs)

    global _graph_client, _stream_client
    _graph_client = GraphClient(config)
    _stream_client = StreamClient(config)


def global_graph_client() -> GraphClient:
    """
    Returns the globally configured GraphQL client. Use
    :class:`~runeq.resources.client.initialize` to configure the client.

    Raises:
        errors.InitializationError: if the library was not initialized.

    """
    if not _graph_client:
        raise INITIALIZATION_ERROR

    return _graph_client


def global_stream_client() -> StreamClient:
    """
    Returns the globally configured Stream API client. Use
    :class:`~runeq.resources.client.initialize` to configure the client.

    Raises:
        errors.InitializationError: if the library was not initialized.


    """
    if not _stream_client:
        raise INITIALIZATION_ERROR

    return _stream_client



"""
Tests for fetching stream data.

"""
from unittest import TestCase, mock

from runeq.config import Config
from runeq.resources.client import StreamClient
from runeq.resources.stream import get_stream_availability, get_stream_data


class TestStreamData(TestCase):
    """
    Unit tests for the data queries in the Stream class.

    """

    def setUp(self):
        """
        Set up mock graph client for testing.

        """
        self.stream_client = StreamClient(
            Config(
                client_key_id='test',
                client_access_key='config'
            )
        )
        self.maxDiff = None

    @mock.patch("runeq.resources.client.requests")
    def test_get_stream_data_csv(self, mock_requests):
        """
        Test get a stream as csv with specific stream_id and optional
        parameters.

        """
        expected_data = """time,acceleration,measurement_duration_ns
2022-07-28T14:26:45.167568Z,0.020525138825178146,20000000
2022-07-28T14:26:45.361596Z,0.020834974944591522,20000000
2022-07-28T14:26:45.361796Z,0.021182861179113388,20000000
2022-07-28T14:26:45.3618588Z,0.022172993049025536,20000000
2022-07-28T14:26:45.3620749Z,0.02356025017797947,20000000
2022-07-28T14:26:45.362149Z,0.024860087782144547,20000000
2022-07-28T14:26:45.36221Z,0.026072751730680466,20000000"""

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.headers = {
            "X-Rune-Next-Page-Token": "foobar"
        }
        mock_response1.text = expected_data

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.headers = {}
        mock_response2.text = expected_data

        mock_requests.get.side_effect = [
            mock_response1,
            mock_response2
        ]

        stream = get_stream_data('test_stream_id', client=self.stream_client,)

        expected = [expected_data, expected_data]
        actual = list(stream)
        self.assertEqual(expected, actual)

    @mock.patch("runeq.resources.client.requests")
    def test_get_stream_data_json(self, mock_requests):
        """
        Test get a stream as json with specific stream_id and optional
        parameters.

        """
        expected_data = {
            "cardinality": 10,
            "data": {
                "acceleration": [
                    0.020525138825178146,
                    0.020834974944591522,
                    0.021182861179113388,
                    0.022172993049025536,
                    0.02356025017797947,
                    0.024860087782144547,
                    0.026072751730680466,
                    0.027338741347193718,
                    0.028795091435313225,
                    0.029819512739777565
                ],
                "measurement_duration_ns": [
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000,
                    20000000
                ],
                "time": [
                    "2022-07-28T14:26:45.167568Z",
                    "2022-07-28T14:26:45.361596Z",
                    "2022-07-28T14:26:45.361796Z",
                    "2022-07-28T14:26:45.3618588Z",
                    "2022-07-28T14:26:45.3620749Z",
                    "2022-07-28T14:26:45.362149Z",
                    "2022-07-28T14:26:45.36221Z",
                    "2022-07-28T14:26:45.362269Z",
                    "2022-07-28T14:26:45.36234Z",
                    "2022-07-28T14:26:45.3624Z"
                ]
            }
        }

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.headers = {
            "X-Rune-Next-Page-Token": "foobar"
        }
        mock_response1.json.return_value = expected_data

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.headers = {}
        mock_response2.json.return_value = expected_data

        mock_requests.get.side_effect = [
            mock_response1,
            mock_response2
        ]

        stream = get_stream_data(
            'test_stream_id',
            format="json",
            client=self.stream_client,
        )

        expected = [expected_data, expected_data]
        actual = list(stream)
        self.assertEqual(expected, actual)

    @mock.patch("runeq.resources.client.requests")
    def test_get_stream_availability_csv(self, mock_requests):
        """
        Test get a stream availability as csv for a specific stream_id.

        """
        expected_data = """time,availability
2022-07-28T14:26:45.167568Z,1
2022-07-28T14:26:45.361596Z,1
2022-07-28T14:26:45.361796Z,0
2022-07-28T14:26:45.3618588Z,0
2022-07-28T14:26:45.3620749Z,0
2022-07-28T14:26:45.362149Z,1
2022-07-28T14:26:45.36221Z,1
"""

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.headers = {
            "X-Rune-Next-Page-Token": "foobar"
        }
        mock_response1.text = expected_data

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.headers = {}
        mock_response2.text = expected_data

        mock_requests.get.side_effect = [
            mock_response1,
            mock_response2
        ]

        availability = get_stream_availability(
            stream_ids='test_stream_id',
            start_time=123,
            end_time=345,
            resolution=300,
            format="csv",
            client=self.stream_client,
        )

        expected = [expected_data, expected_data]
        actual = list(availability)
        self.assertEqual(expected, actual)

    @mock.patch("runeq.resources.client.requests")
    def test_get_stream_availability_json(self, mock_requests):
        """
        Test get a stream availability as json for a specific stream_id.

        """
        expected_data = {
            "data": {
                "time": [
                    "2022-07-28T14:25:00Z",
                    "2022-07-28T14:30:00Z",
                    "2022-07-28T14:35:00Z",
                    "2022-07-28T14:40:00Z",
                    "2022-07-28T14:45:00Z",
                    "2022-07-28T14:50:00Z",
                    "2022-07-28T14:55:00Z",
                    "2022-07-28T15:00:00Z",
                    "2022-07-28T15:05:00Z",
                    "2022-07-28T15:10:00Z"
                ],
                "availability": [
                    1,
                    1,
                    0,
                    0,
                    1,
                    1,
                    0,
                    1,
                    1,
                    1
                ]
            },
            "approx_available_duration_s": 3000,
            "cardinality": 10
        }

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.headers = {
            "X-Rune-Next-Page-Token": "foobar"
        }
        mock_response1.json.return_value = expected_data

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.headers = {}
        mock_response2.json.return_value = expected_data

        mock_requests.get.side_effect = [
            mock_response1,
            mock_response2
        ]

        availability = get_stream_availability(
            stream_ids='test_stream_id',
            start_time=123,
            end_time=345,
            resolution=300,
            format="json",
            client=self.stream_client,
        )

        expected = [expected_data, expected_data]
        actual = list(availability)
        self.assertEqual(expected, actual)

    @mock.patch("runeq.resources.client.requests")
    def test_get_batch_stream_availability(self, mock_requests):
        """
        Test get batch stream availability for multiple stream_ids.

        """
        expected_data = """time,availability
2022-07-28T14:26:45.167568Z,1
2022-07-28T14:26:45.361596Z,1
2022-07-28T14:26:45.361796Z,0
2022-07-28T14:26:45.3618588Z,0
2022-07-28T14:26:45.3620749Z,0
2022-07-28T14:26:45.362149Z,1
2022-07-28T14:26:45.36221Z,1
"""

        # Mock a paginated response
        mock_response1 = mock.Mock()
        mock_response1.ok = True
        mock_response1.headers = {
            "X-Rune-Next-Page-Token": "foobar"
        }
        mock_response1.text = expected_data

        mock_response2 = mock.Mock()
        mock_response2.ok = True
        mock_response2.headers = {}
        mock_response2.text = expected_data

        mock_requests.get.side_effect = [
            mock_response1,
            mock_response2
        ]

        # Must include batch_operation when querying >1 stream
        with self.assertRaisesRegex(
            ValueError,
            "batch_operation must be specified"
        ):
            get_stream_availability(
                stream_ids=['test_stream_id1', 'test_stream_id2'],
                start_time=123,
                end_time=345,
                resolution=300,
                format="csv",
                client=self.stream_client,
            ).__next__()

        availability = get_stream_availability(
            stream_ids=['test_stream_id1', 'test_stream_id2'],
            start_time=123,
            end_time=345,
            resolution=300,
            batch_operation="all",
            client=self.stream_client,
        )

        expected = [expected_data, expected_data]
        actual = list(availability)
        self.assertEqual(expected, actual)



"""
Tests for the V2 SDK Client.

"""
from unittest import TestCase

from runeq import errors
from runeq.resources.client import initialize
from runeq.resources.stream import get_stream_data


class TestInitialize(TestCase):
    """
    Unit tests for the initialization of the user's credentials.

    """

    def test_init_invalid(self):
        """
        Test that access is restricted with no initialization.

        """
        with self.assertRaises(errors.InitializationError):
            get_stream_data("stream_id").__next__()


    def test_init_with_access_keys(self):
        """
        Test that the user is initialized but with invalid access keys.

        """
        initialize(
            access_token_id="foo",
            access_token_secret="bar",
        )

        with self.assertRaisesRegex(errors.APIError, "404 NotFound"):
            get_stream_data("stream_id").__next__()

    def test_init_with_client_keys(self):
        """
        Test that the user is initialized but with invalid client keys.

        """
        initialize(
            client_key_id="foo",
            client_access_key="bar",
        )

        with self.assertRaisesRegex(
            errors.APIError,
            "401 InvalidAuthentication"
        ):
            get_stream_data("stream_id").__next__()



"""
Error classes.

"""


class RuneError(Exception):
    """
    Base class for Rune SDK errors.

    """


class InitializationError(RuneError):
    """
    Rune Initialization Error. Raised when a user has not initialized
    their API credentials.

    """


class APIError(RuneError):
    """
    Rune API Error. Includes details about error type.

    """

    def __init__(self, status_code, details):
        """
        Init with HTTP status code and details
        """
        self.status_code = status_code
        self.details = details

        err_type = 'Error'
        if isinstance(details, dict) and 'type' in details:
            err_type = details['type']

        super().__init__(f'{status_code} {err_type}: {details}')



"""
Fetch metadata about organizations. A research lab or clinical site is
typically represented as an organization.

"""

from typing import Iterable, Optional, Type

from .client import GraphClient, global_graph_client
from .common import ItemBase, ItemSet


class Org(ItemBase):
    """
    Metadata for an organization.

    """

    def __init__(
        self,
        id: str,
        name: str,
        created_at: float,
        **attributes
    ):
        """
        Initialize with metadata.

        Args:
            id: ID of the organization
            name: Human-readable name
            created_at: When the organization was created (unix timestamp)
            **attributes: Other attributes associated with the organization

        """
        norm_id = Org.normalize_id(id)
        self.name = name
        self.created_at = created_at

        super().__init__(
            id=norm_id,
            name=name,
            created_at=created_at,
            **attributes,
        )

    @staticmethod
    def normalize_id(org_id: str) -> str:
        """
        Strip resource prefix and suffix from an org ID (if they exist).

        Args:
            org_id: Organization ID

        """
        if org_id.startswith("org-"):
            org_id = org_id[4:]

        if org_id.endswith(",org"):
            org_id = org_id[:-4]

        return org_id

    @staticmethod
    def denormalize_id(org_id: str) -> str:
        """
        Add resource prefix and suffix to an org ID (if they don't exist).

        This constructs the form of the ID that is used for requests to the
        GraphQL API.

        Args:
            org_id: Organization ID

        """
        if not org_id.startswith("org-"):
            org_id = f'org-{org_id}'

        if not org_id.endswith(",org"):
            org_id = f'{org_id},org'

        return org_id


class OrgSet(ItemSet):
    """
    A collection of Organizations.

    """

    def __init__(self, items: Iterable[Org] = ()):
        """
        Initialize with Orgs.

        """
        super().__init__(items=items)

    @property
    def _item_class(self) -> Type[ItemBase]:
        """
        Instance type of items in this set.

        """
        return Org


def get_org(org_id: str, client: Optional[GraphClient] = None) -> Org:
    """
    Get the org with the specified ID.

    Args:
        org_id: Organization ID
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    client = client or global_graph_client()
    query = '''
        query getOrg($org_id: ID) {
            org (orgId: $org_id) {
                id
                created_at: created
                name: displayName
            }
        }
    '''

    full_org_id = Org.denormalize_id(org_id)

    result = client.execute(
        statement=query,
        org_id=full_org_id
    )

    org_attrs = result["org"]
    return Org(
        id=org_attrs["id"],
        created_at=org_attrs.get("created_at"),
        name=org_attrs.get("name"),
    )


def get_orgs(client: Optional[GraphClient] = None) -> OrgSet:
    """
    Get all the organizations that the current user is a member of.

    Args:
        client: If specified, this client is used to fetch metadata from the
            API. Otherwise, the global GraphClient is used.

    """
    client = client or global_graph_client()
    query = '''
        query getOrgMemberships($cursor: Cursor) {
            user {
                membershipList(cursor: $cursor) {
                    pageInfo {
                        endCursor
                    }
                    memberships {
                        org {
                            id
                            created_at: created
                            name: displayName
                        }
                    }
                }
            }
        }
    '''

    next_cursor = None
    org_set = OrgSet()

    # Use cursor to page through all org memberships
    while True:
        result = client.execute(
            statement=query,
            cursor=next_cursor
        )

        membership_list = result.get("user", {}).get("membershipList", {})

        for membership in membership_list.get("memberships", []):
            org_attrs = membership["org"]
            org = Org(
                id=org_attrs["id"],
                created_at=org_attrs.get("created_at"),
                name=org_attrs.get("name"),
            )
            org_set.add(org)

        # endCursor is None when there are no more pages of data
        next_cursor = membership_list.get("pageInfo", {}).get("endCursor")
        if not next_cursor:
            break

    return org_set
